#include <windows.h>

#include <stdio.h>
#include <time.h>

#include "types.h"
#include "bitops.h"
#include "buf.h"
#include "vfs.h"
#include "dfs.h"

#define K 1024

#define SECTORSIZE       512

#define BLOCKSIZE        4096
#define INODE_RATIO      4096

char bootsect[SECTORSIZE];

char *command;
char *devname;
char *bootfile;
char *filedir;
int devsize;

int get_tick_count()
{
  return (int) GetTickCount();
}

void panic(char *reason)
{
  printf("panic: %s\n", reason);
  exit(1);
}

int dev_read(devno_t devno, void *buffer, size_t count, blkno_t blkno)
{
  DWORD bytes;

  //printf("read block %d, %d bytes\n", blkno, count);

  if (SetFilePointer((HANDLE) devno, blkno * SECTORSIZE, NULL, FILE_BEGIN) == -1) panic("unable to set file pointer");
  if (!ReadFile((HANDLE) devno, buffer, count, &bytes, NULL)) panic("error reading from device");
  return count;
}

int dev_write(devno_t devno, void *buffer, size_t count, blkno_t blkno)
{
  DWORD bytes;

  //printf("write block %d, %d bytes\n", blkno, count);

  if (SetFilePointer((HANDLE) devno, blkno * SECTORSIZE, NULL, FILE_BEGIN) == -1) panic("unable to set file pointer");
  if (!WriteFile((HANDLE) devno, buffer, count, &bytes, NULL)) panic("error writing to device");
  return count;
}

unsigned int dev_getsize(devno_t devno)
{
  unsigned int size;

  //size = GetFileSize((HANDLE) devno, NULL) / SECTORSIZE;
  size = devsize;
  printf("device size is %d sectors (%dMB)\n", size, size / 2048);
  return size;
}

void create_device(char *devname, int devsize)
{
  DWORD bytes;
  HANDLE hdev;
  char sector[SECTORSIZE];
  int i;

  hdev = CreateFile(devname, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
  if (hdev == INVALID_HANDLE_VALUE) panic("unable to create device file");

  memset(sector, 0, SECTORSIZE);
  for (i = 0; i < devsize; i++)
  {
    if (!WriteFile(hdev, sector, SECTORSIZE, &bytes, NULL)) panic("error writing to device");
  }

  CloseHandle(hdev);
}

void clear_device(HANDLE hdev, int devsize)
{
  DWORD bytes;
  char sector[SECTORSIZE];
  int i;

  memset(sector, 0, SECTORSIZE);
  for (i = 0; i < devsize; i++)
  {
    if (!WriteFile(hdev, sector, SECTORSIZE, &bytes, NULL)) panic("error writing to device");
  }
}

void read_boot_sector(char *name)
{
  HANDLE hfile;
  DWORD bytes;

  hfile = CreateFile(name, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
  if (hfile == INVALID_HANDLE_VALUE) panic("unable to read boot sector");

  ReadFile(hfile, bootsect, SECTORSIZE, &bytes, NULL);

  CloseHandle(hfile);
}

void make_directory(char *dirname)
{
  mkdir(dirname);
}

void transfer_file(char *dstfn, char *srcfn)
{
  HANDLE hfile;
  DWORD size;
  size_t count;

  struct file *file;
  char buf[4096];

  hfile = CreateFile(srcfn, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
  size = GetFileSize(hfile, NULL);

  file = open(dstfn, O_CREAT);

  printf("%s -> %s (%d bytes)\n", srcfn, dstfn, size);

  ReadFile(hfile, buf, 4096, &count, NULL);
  while (count > 0)
  {
    write(file, buf, count);
    ReadFile(hfile, buf, 4096, &count, NULL);
  }
  
  close(file);
  CloseHandle(hfile);
}

void transfer_files(char *dstdir, char *srcdir)
{
  WIN32_FIND_DATA finddata;
  HANDLE hfind;
  BOOL more;
  char srcfn[255];
  char dstfn[255];

  strcpy(srcfn, srcdir);
  strcat(srcfn, "\\*");

  hfind = FindFirstFile(srcfn, &finddata);
  more = hfind != INVALID_HANDLE_VALUE;
  while (more)
  {
    strcpy(srcfn, srcdir);
    strcat(srcfn, "\\");
    strcat(srcfn, finddata.cFileName);

    strcpy(dstfn, dstdir);
    strcat(dstfn, "/");
    strcat(dstfn, finddata.cFileName);

    if (finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
      if (strcmp(finddata.cFileName, ".") != 0 && strcmp(finddata.cFileName, "..") != 0)
      {
	make_directory(dstfn);
	transfer_files(dstfn, srcfn);
      }
    }
    else
      transfer_file(dstfn, srcfn);

    more = FindNextFile(hfind, &finddata);
  }

  FindClose(hfind);
}

void shell();

void create_fs()
{
  HANDLE hdev;

  // Register dfs filesystem
  dfs_init();

  // Read boot sector
  read_boot_sector(bootfile);

  // Create device file
  if (strncmp(devname, "\\\\.\\", 4) != 0) 
  {
    printf("Creating device file %s\n", devname);
    create_device(devname, devsize);
  }

  // Open device file
  hdev = CreateFile(devname, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH | FILE_FLAG_NO_BUFFERING, NULL);
  if (hdev == INVALID_HANDLE_VALUE) 
  {
    printf("Error %d opening device %s\n", GetLastError(), devname);
    panic("unable to open device");
  }

  // Clear device
  if (strncmp(devname, "\\\\.\\", 4) == 0) 
  {
    printf("Clearing device %s\n", devname);
    clear_device(hdev, devsize);
  }

  // Write boot sector
  dev_write((devno_t) hdev, bootsect, SECTORSIZE, 0);

  // Create filesystem
  format((devno_t) hdev, "dfs", NULL);

  // Initialize the file system
  mount("dfs", "/", (devno_t) hdev, NULL);

  // Copy files to device
  transfer_files("", filedir);

  // Close file system
  unmount_all();

  // Close device file
  CloseHandle(hdev);
}

void transfer()
{
  HANDLE hdev;

  // Register dfs filesystem
  dfs_init();

  // Open device file
  hdev = CreateFile(devname, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH | FILE_FLAG_NO_BUFFERING, NULL);
  if (hdev == INVALID_HANDLE_VALUE) 
  {
    printf("Error %d opening device %s\n", GetLastError(), devname);
    panic("unable to open device");
  }

  // Initialize the file system
  mount("dfs", "/", (devno_t) hdev, NULL);

  // Copy file to device
  transfer_file(bootfile, filedir);

  // Close file system
  unmount_all();

  // Close device file
  CloseHandle(hdev);
}

int main(int argc, char **argv)
{
  // Parse arguments
  if (argc < 2)
  {
    printf("usage: mkdfs <command> <devname> <source> [<bootfile>|<dest>] <devsize> (devsize in KB)\n");
    printf("             command = create | transfer | shell\n");
    return 1;
  }

  command = argv[1];
  devname = argv[2];
  filedir = argv[3];
  bootfile = argv[4];
  devsize = atoi(argv[5]) * 2; // two sectors per KB

  if (strcmp(command, "shell") == 0)
    shell();
  else if (strcmp(command, "create") == 0)
    create_fs();
  else if (strcmp(command, "transfer") == 0)
    transfer();
  else
    printf("%s: unknown command\n", command);

  return 0;
}

void list_file(char *filename)
{
  size_t count;
  char buf[4096];
  struct file *file;

  file = open(filename, 0);
  if (!file)
  {
    printf("%s: file not found\n", filename);
    return;
  }

  while ((count = read(file, buf, 4096)) > 0)
  {
    fwrite(buf, 1, count, stdout);
  }

  close(file);
}

void copy_file(char *srcfn, char *dstfn)
{
  size_t count;
  char buf[4096];
  struct file *f1;
  struct file *f2;

  f1 = open(srcfn, 0);
  if (!f1)
  {
    printf("%s: file not found\n", srcfn);
    return;
  }

  f2 = open(dstfn, O_CREAT);
  if (!f2)
  {
    close(f1);
    printf("%s: unable to create file\n", dstfn);
    return;
  }

  while ((count = read(f1, buf, 4096)) > 0)
  {
    write(f2, buf, count);
  }

  close(f1);
  close(f2);
}

void list_dir(char *dirname)
{
  struct file *dir;
  struct dirent dirp;
  struct stat buf;
  struct tm *tm;
  char path[MAXPATH];

  dir = opendir(dirname);
  if (!dir)
  {
    printf("%s: directory not found\n", dirname);
    return;
  }

  while (readdir(dir, &dirp, 1) > 0)
  {
    strcpy(path, dirname);
    strcat(path, "/");
    strcat(path, dirp.name);

    stat(path, &buf);
    tm = gmtime(&buf.ctime);

    printf("%8d %8d %02d/%02d/%04d %02d:%02d:%02d ", buf.ino, buf.quad.size_low, tm->tm_mday, tm->tm_mon, tm->tm_year + 1900, tm->tm_hour, tm->tm_min, tm->tm_sec);
    if (buf.mode & FS_DIRECTORY) 
      printf("[%s]", dirp.name);
    else
      printf("%s", dirp.name);

    printf("\n");
  }

  close(dir);
}

void remove_file(char *filename)
{
  if (xunlink(filename) < 0)
  {
    printf("%s: file not found\n", filename);
    return;
  }
}

void make_dir(char *filename)
{
  if (mkdir(filename) < 0)
  {
    printf("%s: cannot make directory\n", filename);
    return;
  }
}

void remove_dir(char *filename)
{
  if (rmdir(filename) < 0)
  {
    printf("%s: cannot delete directory\n", filename);
    return;
  }
}

void disk_usage(char *filename)
{
  struct fs *fs;
  struct filsys *filsys;

  fs = fslookup(filename, NULL);
  if (!fs)
  {
    printf("%s: unknown file system\n", filename);
  }

  filsys = (struct filsys *) fs->data;
  printf("block size %d\n", filsys->blocksize);
  printf("blocks %d\n", filsys->super->block_count);
  printf("inodes %d\n", filsys->super->inode_count);
  printf("free blocks %d\n", filsys->super->free_block_count);
  printf("free inodes %d\n", filsys->super->free_inode_count);

  printf("%dKB used, %dKB free, %dKB total\n", 
    filsys->blocksize * (filsys->super->block_count - filsys->super->free_block_count) / 1024, 
    filsys->blocksize * filsys->super->free_block_count / 1024, 
    filsys->blocksize * filsys->super->block_count / 1024);
}

static void test_read_file(char *filename)
{
  int count;
  struct file *file;
  char *data;
  int start;
  int time;
  int bytes;

  file = open(filename, 0);
  if (!file)
  {
    printf("%s: file not found\n", filename);
    return;
  }

  data = malloc(64 * K);

  bytes = 0;
  start = get_tick_count();
  while ((count = read(file, data, 64 * K)) > 0)
  {
    bytes += count;
  }
  time = (get_tick_count() - start) * 10;
  printf("%s: read %dKB in %d ms, %dKB/s\n", filename, bytes / K, time, bytes / time);
  
  free(data);

  if (count < 0) printf("%s: error reading file\n", filename);

  close(file);
}

static void test_write_file(char *filename, int size)
{
  int count;
  struct file *file;
  char *data;
  int start;
  int time;
  int bytes;

  file = open(filename, O_CREAT);
  if (!file)
  {
    printf("%s: error creating file\n", filename);
    return;
  }

  data = malloc(64 * K);

  bytes = 0;
  start = get_tick_count();
  while (bytes < size)
  {
    if ((count = write(file, data, 64 * K)) <= 0)
    {
      printf("%s: error writing file\n", filename);
      break;
    }

    bytes += count;
  }
  time = (get_tick_count() - start) * 10;
  printf("%s: wrote %dKB in %d ms, %dKB/s\n", filename, bytes / K, time, bytes / time);
  
  free(data);

  close(file);
}

void shell()
{
  HANDLE hdev;
  char cmd[256];
  char *arg;
  char *arg2;

  // Register dfs filesystem
  dfs_init();

  // Open device file
  hdev = CreateFile(devname, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, 0, NULL);
  if (hdev == INVALID_HANDLE_VALUE) panic("unable to open device");

  // Initialize the file system
  mount("dfs", "/", (devno_t) hdev, NULL);

  while (1)
  {
    printf("> ");
    gets(cmd);

    arg = cmd;
    while (*arg && *arg != ' ') arg++;
    while (*arg == ' ') *arg++ = 0;

    arg2 = arg;
    while (*arg2 && *arg2 != ' ') arg2++;
    while (*arg2 == ' ') *arg2++ = 0;

    if (strcmp(cmd, "exit") == 0) 
      break;
    else if (strcmp(cmd, "ls") == 0)
      list_dir(arg);
    else if (strcmp(cmd, "cat") == 0)
      list_file(arg);
    else if (strcmp(cmd, "cp") == 0)
      copy_file(arg, arg2);
    else if (strcmp(cmd, "rm") == 0)
      remove_file(arg);
    else if (strcmp(cmd, "mkdir") == 0)
      make_dir(arg);
    else if (strcmp(cmd, "rmdir") == 0)
      remove_dir(arg);
    else if (strcmp(cmd, "du") == 0)
      disk_usage(arg);
    else if (strcmp(cmd, "read") == 0)
      test_read_file(arg);
    else if (strcmp(cmd, "write") == 0)
      test_write_file(arg, atoi(arg2) * K);
    else if (*cmd)
      printf("%s: unknown command\n", cmd);
  }

  // Close file system
  unmount_all();

  // Close device file
  CloseHandle(hdev);
}
