//
// hd.c
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// Harddisk driver
//

#include <os/krnl.h>

#define SECTORSIZE              512

#define HD_CONTROLLERS		2
#define HD_DRIVES		4
#define HD_PARTITIONS		4

#define MAX_PRDS                (PAGESIZE / 8)
#define MAX_DMA_XFER_SIZE       ((MAX_PRDS - 1) * PAGESIZE)

#define HDC0_IOBASE		0x01F0
#define HDC1_IOBASE		0x0170

#define HDC0_IRQ                14
#define HDC1_IRQ                15

#define HD0_DRVSEL              0xA0
#define HD1_DRVSEL              0xB0

//
// Controller registers
//

#define HDC_DATA		0x0000
#define HDC_ERR			0x0001
#define HDC_FEATURE		0x0001
#define HDC_SECTORCNT		0x0002
#define HDC_SECTOR		0x0003
#define HDC_TRACKLSB		0x0004
#define HDC_TRACKMSB		0x0005
#define HDC_DRVHD		0x0006
#define HDC_STATUS		0x0007
#define HDC_COMMAND		0x0007
#define HDC_DEVCTRL		0x0008
#define HDC_ALT_STATUS		0x0206
#define HDC_CONTROL		0x0206

//
// Controller commands
//

#define HDCTL_RST		0x04   // Controller reset
#define HDCTL_4BITHD		0x08   // Use 4 bits for head id

//
// Drive commands
//

#define HDCMD_NULL		0x00
#define HDCMD_IDENTIFY		0xEC
#define HDCMD_RESET		0x10
#define HDCMD_READ		0x20
#define HDCMD_WRITE		0x30
#define HDCMD_READ_DMA          0xC8
#define HDCMD_WRITE_DMA         0xCA

//
// Controller status
//

#define HDCS_ERR		0x01   // Error
#define HDCS_IDX		0x02   // Index
#define HDCS_CORR		0x04   // Corrected data
#define HDCS_DRQ		0x08   // Data request
#define HDCS_DSC		0x10   // Drive seek complete
#define HDCS_DWF		0x20   // Drive write fault
#define HDCS_DRDY		0x40   // Drive ready
#define HDCS_BSY		0x80   // Controller busy

//
// Device control 
//

#define HDDC_HD15               0x08  // Bit should always be set to one
#define HDDC_SRST               0x04  // Soft reset
#define HDDC_NIEN               0x02  // Disable interrupts

//
// Controller error conditions 
//

#define HDCE_AMNF		0x01   // Address mark not found
#define HDCE_TK0NF		0x02   // Track 0 not found
#define HDCE_ABRT		0x04   // Abort
#define HDCE_MCR		0x08   // Media change requested
#define HDCE_IDNF		0x10   // Sector id not found
#define HDCE_MC			0x20   // Media change
#define HDCE_UNC		0x40   // Uncorrectable data error
#define HDCE_BBK		0x80   // Bad block

//
// Timeouts (in ms)
//

#define HDTIMEOUT_DRDY		5000
#define HDTIMEOUT_DRQ		5000
#define HDTIMEOUT_CMD		2000

//
// Buffer type
//

#define HDTYPE_SINGLE           1      // Single port, single sector buf
#define HDTYPE_DUAL             2      // Dual port, multiple sector buf
#define HDTYPE_DUAL_CACHE       3      // Above plus track cache

//
// Transfer type
//

#define HD_XFER_IDLE            0
#define HD_XFER_READ            1
#define HD_XFER_WRITE           2
#define HD_XFER_DMA             3
#define HD_XFER_IGNORE          4

//
// Bus master registers
//

#define BM_COMMAND_REG    0            // Offset to command reg
#define BM_STATUS_REG     2            // Offset to status reg
#define BM_PRD_ADDR_LOW   4            // Offset to PRD addr reg low 16 bits
#define BM_PRD_ADDR_HIGH  6            // Offset to PRD addr reg high 16 bits

//
// Bus master command register flags
//

#define BM_CR_MASK_READ    0x00        // Read from memory
#define BM_CR_MASK_WRITE   0x08        // Write to memory
#define BM_CR_MASK_START   0x01        // Start transfer
#define BM_CR_MASK_STOP    0x00        // Stop transfer

//
// Bus master status register flags
//

#define BM_SR_MASK_SIMPLEX 0x80        // Simplex only
#define BM_SR_MASK_DRV1    0x40        // Drive 1 can do dma
#define BM_SR_MASK_DRV0    0x20        // Drive 0 can do dma
#define BM_SR_MASK_INT     0x04        // INTRQ signal asserted
#define BM_SR_MASK_ERR     0x02        // Error
#define BM_SR_MASK_ACT     0x01        // Active

//
// Parameters returned by read drive parameters command

struct hdparam 
{
  // Drive information
  short config;		               // General configuration bits
  unsigned short cylinders;	       // Cylinders
  short reserved;
  unsigned short heads;		       // Heads
  short unfbytespertrk;	               // Unformatted bytes/track
  short unfbytes;		       // Unformatted bytes/sector
  unsigned short sectors;	       // Sectors per track
  short vendorunique[3];

  // Controller information
  char serial[20];		       // Serial number
  short buffertype;		       // Buffer type
  short buffersize;		       // Buffer size, in 512-byte units
  short necc;			       // ECC bytes appended
  char rev[8];		               // Firmware revision
  char model[40];		       // Model name
  short nsecperint;		       // Sectors per interrupt
  short usedmovsd;		       // Can use double word read/write?
  char pad[418];
};

struct hd;

struct prd
{
  void *addr;
  int len;
};

struct hdc 
{
  struct mutex lock;	               // Controller mutex
  struct event ready;	               // Controller interrupt event
  struct dpc xfer_dpc;                 // DPC for data transfer
  
  int status;                          // Controller status
  int bmstat;                          // Busmaster status

  int iobase;	                       // I/O port registers base address
  int irq;                             // IRQ for controller
  int bmregbase;                       // Busmaster register base

  char *bufp;                          // Buffer pointer for next transfer
  int nsects;                          // Number of sectors left to transfer
  int result;                          // Result of transfer
  int dir;                             // Transfer direction
  struct hd *active;                   // Active drive for transfer

  struct prd *prds;                    // PRD list for DMA transfer
  unsigned long prds_phys;             // Physical address of PRD list
};

struct partition
{
  devno_t dev;
  unsigned int start;
  unsigned int len;
  unsigned short bootid;
  unsigned short systid;
};

struct hd 
{
  struct hdc *hdc;		        // Controller
  struct hdparam param;		        // Drive parameter block
  int drvsel;                           // Drive select on controller
  int use32bits;                        // Use 32 bit transfers

  // Geometry
  unsigned int blks;		        // Number of blocks on drive
  unsigned int size;		        // Size in MB

  unsigned int cyls;	                // Number of cylinders
  unsigned int heads;		        // Number of heads
  unsigned int sectors;                 // Sectors per track

  struct partition parts[HD_PARTITIONS]; // Partition info
};

struct pci_dev *idedev;

static struct hdc hdctab[HD_CONTROLLERS];
static struct hd hdtab[HD_DRIVES];

static char *hd0parts[HD_PARTITIONS] = {"hd0a", "hd0b", "hd0c", "hd0d"};
static char *hd1parts[HD_PARTITIONS] = {"hd1a", "hd1b", "hd1c", "hd1d"};
static char *hd2parts[HD_PARTITIONS] = {"hd2a", "hd2b", "hd2c", "hd2d"};
static char *hd3parts[HD_PARTITIONS] = {"hd3a", "hd3b", "hd3c", "hd3d"};

static void insw(int port, void *buf, int count)
{
  __asm
  {
    mov edx, port
    mov edi, buf
    mov ecx, count
    rep insw
  }
}

static void outsw(int port, void *buf, int count)
{
  __asm
  {
    mov edx, port
    mov esi, buf
    mov ecx, count
    rep outsw
  }
}

static void insd(int port, void *buf, int count)
{
  __asm
  {
    mov edx, port
    mov edi, buf
    mov ecx, count
    rep insd
  }
}

static void outsd(int port, void *buf, int count)
{
  __asm
  {
    mov edx, port
    mov esi, buf
    mov ecx, count
    rep outsd
  }
}

static void hd_error(char *func, unsigned char error)
{
  kprintf("%s: ", func);
  if (error & HDCE_BBK)   kprintf("bad block  ");
  if (error & HDCE_UNC)   kprintf("uncorrectable data  ");
  if (error & HDCE_MC)    kprintf("media change  ");
  if (error & HDCE_IDNF)  kprintf("id not found  ");
  if (error & HDCE_MCR)   kprintf("media change requested  ");
  if (error & HDCE_ABRT)  kprintf("abort  ");
  if (error & HDCE_TK0NF) kprintf("track 0 not found  ");
  if (error & HDCE_AMNF)  kprintf("address mark not found  ");
  kprintf("\n");
}

static int hd_wait(struct hdc *hdc, unsigned char mask, unsigned int timeout)
{
  unsigned int start;
  unsigned char status;

  start = get_tick_count();
  while (1)
  {
    status = _inp((unsigned short) (hdc->iobase + HDC_ALT_STATUS));
    if (status & HDCS_ERR) 
    {
       unsigned char error;
 
       error = _inp((unsigned short) (hdc->iobase + HDC_ERR));
       hd_error("hdwait", error);

       return error;
    }

    if (!(status & HDCS_BSY) && ((status & mask) == mask)) return 0;
    if (get_tick_count() - start >= timeout) return -1;

    yield();
  }
}

static void hd_setup_transfer(struct hd *hd, blkno_t blkno, int nsects)
{
  unsigned int track;
  unsigned int head;
  unsigned int sector;

  track = blkno / (hd->heads * hd->sectors);
  head = (blkno / hd->sectors) % hd->heads;
  sector = blkno % hd->sectors + 1;

  _outp((unsigned short) (hd->hdc->iobase + HDC_SECTORCNT), nsects);
  _outp((unsigned short) (hd->hdc->iobase + HDC_SECTOR), (unsigned char) sector);
  _outp((unsigned short) (hd->hdc->iobase + HDC_TRACKLSB), (unsigned char) track);
  _outp((unsigned short) (hd->hdc->iobase + HDC_TRACKMSB), (unsigned char) (track >> 8));
  _outp((unsigned short) (hd->hdc->iobase + HDC_DRVHD), ((unsigned char) head & 0x0F) | hd->drvsel);
}

static void pio_read_sect(struct hd *hd, char *buffer)
{
  struct hdc *hdc = hd->hdc;

  if (hd->use32bits)
    insd(hdc->iobase + HDC_DATA, buffer, SECTORSIZE / 4);
  else
    insw(hdc->iobase + HDC_DATA, buffer, SECTORSIZE / 2);
}

static void pio_write_sect(struct hd *hd, char *buffer)
{
  struct hdc *hdc = hd->hdc;

  if (hd->use32bits)
    outsd(hdc->iobase + HDC_DATA, buffer, SECTORSIZE / 4);
  else
    outsw(hdc->iobase + HDC_DATA, buffer, SECTORSIZE / 2);
}

static void setup_dma_transfer(struct hdc *hdc, char *buffer, int count)
{
  int i;
  int len;
  char *next;

  i = 0;
  next = (char *) ((unsigned long) buffer & ~PAGESIZE) + PAGESIZE;
  while (1)
  {
    if (i == MAX_PRDS) panic("hd dma transfer too large");

    hdc->prds[i].addr = virt2phys(buffer);
    len = next - buffer;
    if (count > len)
    {
      hdc->prds[i].len = len;
      count -= len;
      buffer = next;
      next += PAGESIZE;
      i++;
    }
    else
    {
      hdc->prds[i].len = count | 0x80000000;
      break;
    }
  }

  _outpw((unsigned short) (hdc->bmregbase + BM_PRD_ADDR_LOW), (unsigned short) (hdc->prds_phys & 0xFFFF));
  _outpw((unsigned short) (hdc->bmregbase + BM_PRD_ADDR_HIGH), (unsigned short) (hdc->prds_phys >> 16));
}

static int hd_identify(struct hd *hd)
{
  int result;

  // Ignore interrupt for identify command
  hd->hdc->dir = HD_XFER_IGNORE;

  // Issue read drive parameters command
  _outp(hd->hdc->iobase + HDC_DRVHD, hd->drvsel);
  _outp(hd->hdc->iobase + HDC_COMMAND, HDCMD_IDENTIFY);

  // Wait for data ready
  if (!(_inp((unsigned short) (hd->hdc->iobase + HDC_ALT_STATUS)) & HDCS_DRQ))
  {
    result = hd_wait(hd->hdc, HDCS_DRQ, HDTIMEOUT_DRQ);
    if (result != 0) return result;
  }

  // Read parameter data */
  insw(hd->hdc->iobase + HDC_DATA, &(hd->param), SECTORSIZE / 2);

  /* Fill in drive parameters */
  hd->cyls = hd->param.cylinders;
  hd->heads = hd->param.heads;
  hd->sectors = hd->param.sectors;
  hd->blks = hd->cyls * hd->heads * hd->sectors;
  hd->size = hd->blks / (M / SECTORSIZE);
  hd->use32bits = hd->param.usedmovsd != 0;

  // Read status
  hd->hdc->status = _inp((unsigned short) (hd->hdc->iobase + HDC_STATUS));

  return 0;
}

static int hd_read_mbr(struct hd *hd, void *buffer)
{
  struct hdc *hdc;
  int result;

  hdc = hd->hdc;

  // Ignore interrupt
  hd->hdc->dir = HD_XFER_IGNORE;

  // Issue read sectors command
  result = hd_wait(hdc, HDCS_DRDY, HDTIMEOUT_DRDY);
  if (result != 0) 
  {
    kprintf("hd_read: no drdy (%d)\n", result);
    return -1;
  }

  hd_setup_transfer(hd, 0, 1);
  _outp((unsigned short) (hdc->iobase + HDC_COMMAND), HDCMD_READ);

  // Wait for data ready
  if (!(_inp((unsigned short)(hdc->iobase + HDC_ALT_STATUS)) & HDCS_DRQ))
  {
    result = hd_wait(hdc, HDCS_DRQ, HDTIMEOUT_DRQ);
    if (result != 0) return -1;
  }

  // Read sector data
  pio_read_sect(hd, (char *) buffer);

  // Check status
  hdc->status = _inp((unsigned short) (hdc->iobase + HDC_STATUS));
  if (hdc->status & HDCS_ERR) return -1;

  return SECTORSIZE;
}

static int hd_ioctl(struct dev *dev, int cmd, void *args)
{
  struct hd *hd = (struct hd *) dev->privdata;

  switch (cmd)
  {
    case IOCTL_GETDEVSIZE:
      return hd->blks;

    case IOCTL_GETBLKSIZE:
      return SECTORSIZE;
  }

  return -1;
}

static int hd_read_poll(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct hd *hd;
  struct hdc *hdc;
  char *hdbuf;
  int i;
  int result;
  int nsects;

  if (count == 0) return 0;
  hd = (struct hd *) dev->privdata;
  hdc = hd->hdc;
  nsects = count / SECTORSIZE;
  wait_for_object(&hdc->lock);

  // Issue read sectors command
  result = hd_wait(hdc, HDCS_DRDY, HDTIMEOUT_DRDY);
  if (result != 0) 
  {
    kprintf("hd_read: no drdy (%d)\n", result);
    release_mutex(&hdc->lock);
    return -1;
  }

  hd_setup_transfer(hd, blkno, nsects);
  _outp((unsigned short) (hdc->iobase + HDC_COMMAND), HDCMD_READ);

  hdbuf = (char *) buffer;
  for (i = 0; i < nsects; i++) 
  {
    // Wait for data ready
    if (!(_inp((unsigned short)(hdc->iobase + HDC_ALT_STATUS)) & HDCS_DRQ))
    {
      result = hd_wait(hdc, HDCS_DRQ, HDTIMEOUT_DRQ);
      if (result != 0)
      {
	kprintf("hd_read: no drq (%d)\n", result);
        release_mutex(&hdc->lock);
        return -1;
      }
    }

    // Read sector data
    pio_read_sect(hd, hdbuf);

    // Check status
    hdc->status = _inp((unsigned short) (hdc->iobase + HDC_STATUS));
    if (hdc->status & HDCS_ERR)
    {
      kprintf("hd_read: read error (%d)\n", hdc->status);
      release_mutex(&hdc->lock);
      return -1;
    }

    hdbuf += SECTORSIZE;
  }

  release_mutex(&hdc->lock);
  return count;
}

static int hd_write_poll(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct hd *hd;
  struct hdc *hdc;
  char *hdbuf;
  int i;
  int result;
  int nsects;

  if (count == 0) return 0;
  hd = (struct hd *) dev->privdata;
  hdc = hd->hdc;
  nsects = count / SECTORSIZE;
  wait_for_object(&hdc->lock);

  // Issue write sectors command
  result = hd_wait(hdc, HDCS_DRDY, HDTIMEOUT_DRDY);
  if (result != 0) 
  {
    kprintf("hd_write: no drdy (%d)\n", result);
    release_mutex(&hdc->lock);
    return -1;
  }

  hd_setup_transfer(hd, blkno, nsects);
  _outp((unsigned short) (hdc->iobase + HDC_COMMAND), HDCMD_WRITE);

  hdbuf = (char *) buffer;
  for (i = 0; i < nsects; i++) 
  {
    // Wait for data ready
    if (!(_inp((unsigned short)(hdc->iobase + HDC_ALT_STATUS)) & HDCS_DRQ))
    {
      result = hd_wait(hdc, HDCS_DRQ, HDTIMEOUT_DRQ);
      if (result != 0)
      {
	kprintf("hd_write: no drq (%d)\n", result);
        release_mutex(&hdc->lock);
        return -1;
      }
    }

    // Write sector data
    pio_write_sect(hd, hdbuf);

    hdbuf += SECTORSIZE;
  }

  release_mutex(&hdc->lock);
  return count;
}

static int hd_read_intr(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct hd *hd;
  struct hdc *hdc;
  int nsects;
  int result;

  if (count == 0) return 0;
  hd = (struct hd *) dev->privdata;
  hdc = hd->hdc;
  nsects = count / SECTORSIZE;
  wait_for_object(&hdc->lock);

  // Wait for controller ready
  result = hd_wait(hdc, HDCS_DRDY, HDTIMEOUT_DRDY);
  if (result != 0) 
  {
    kprintf("hd_read: no drdy (%d)\n", result);
    release_mutex(&hdc->lock);
    return -1;
  }

  // Prepare transfer
  hdc->bufp = (char *) buffer;
  hdc->nsects = nsects;
  hdc->result = 0;
  hdc->dir = HD_XFER_READ;
  hdc->active = hd;

  hd_setup_transfer(hd, blkno, nsects);
  _outp((unsigned short) (hdc->iobase + HDC_COMMAND), HDCMD_READ);

  // Wait until data read
  wait_for_object(&hdc->ready);

  // Cleanup
  hdc->dir = HD_XFER_IDLE;
  hdc->active = NULL;
  result = hdc->result;
  release_mutex(&hdc->lock);
  return result == 0 ? count : -1;
}

static int hd_write_intr(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct hd *hd;
  struct hdc *hdc;
  int nsects;
  int result;

  if (count == 0) return 0;
  hd = (struct hd *) dev->privdata;
  hdc = hd->hdc;
  nsects = count / SECTORSIZE;
  wait_for_object(&hdc->lock);

  // Wait for controller ready
  result = hd_wait(hdc, HDCS_DRDY, HDTIMEOUT_DRDY);
  if (result != 0) 
  {
    kprintf("hd_write: no drdy (%d)\n", result);
    release_mutex(&hdc->lock);
    return -1;
  }

  // Prepare transfer
  hdc->bufp = (char *) buffer;
  hdc->nsects = nsects;
  hdc->result = 0;
  hdc->dir = HD_XFER_WRITE;
  hdc->active = hd;

  hd_setup_transfer(hd, blkno, nsects);
  _outp((unsigned short) (hdc->iobase + HDC_COMMAND), HDCMD_WRITE);

  // Write first sector
  pio_write_sect(hd, hdc->bufp);

  // Wait until data written
  wait_for_object(&hdc->ready);

  // Cleanup
  hdc->dir = HD_XFER_IDLE;
  hdc->active = NULL;
  result = hdc->result;
  release_mutex(&hdc->lock);
  return result == 0 ? count : -1;
}

static int hd_read_dma(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct hd *hd;
  struct hdc *hdc;
  int nsects;
  int result;

  if (count == 0) return 0;
  if (count > MAX_DMA_XFER_SIZE) return hd_read_intr(dev, buffer, count, blkno);

  hd = (struct hd *) dev->privdata;
  hdc = hd->hdc;
  nsects = count / SECTORSIZE;
  wait_for_object(&hdc->lock);

  // Wait for controller ready
  result = hd_wait(hdc, HDCS_DRDY, HDTIMEOUT_DRDY);
  if (result != 0) 
  {
    kprintf("hd_read_dma: no drdy (%d)\n", result);
    release_mutex(&hdc->lock);
    return -1;
  }

  // Setup DMA
  _outp((unsigned short) (hdc->bmregbase + BM_STATUS_REG), hdc->bmstat | BM_SR_MASK_INT | BM_SR_MASK_ERR);
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_STOP);
  setup_dma_transfer(hdc, buffer, count);
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_WRITE);

  // Prepare transfer
  result = 0;
  hdc->dir = HD_XFER_DMA;
  hdc->active = hd;

  hd_setup_transfer(hd, blkno, nsects);
  _outp((unsigned short) (hdc->iobase + HDC_COMMAND), HDCMD_READ_DMA);
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_WRITE | BM_CR_MASK_START);

  // Wait for interrupt
  wait_for_object(&hdc->ready);

  // Now wait for the PCI BM DMA channel to flush the data
  while (1)
  {
    if (_inp((unsigned short) (hdc->bmregbase + BM_STATUS_REG)) & BM_SR_MASK_INT) break;
  }

  // Check status
  hdc->status = _inp((unsigned short) (hdc->iobase + HDC_STATUS));
  if (hdc->status & HDCS_ERR)
  {
    kprintf("hd: dma read error (%d)\n", hdc->status);
    result = -1;
  }

  // Stop DMA channel
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_STOP);

  // Cleanup
  hdc->dir = HD_XFER_IDLE;
  hdc->active = NULL;
  release_mutex(&hdc->lock);
  return result == 0 ? count : -1;
}

static int hd_write_dma(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct hd *hd;
  struct hdc *hdc;
  int nsects;
  int result;

  if (count == 0) return 0;
  if (count > MAX_DMA_XFER_SIZE) return hd_read_intr(dev, buffer, count, blkno);

  hd = (struct hd *) dev->privdata;
  hdc = hd->hdc;
  nsects = count / SECTORSIZE;
  wait_for_object(&hdc->lock);

  // Wait for controller ready
  result = hd_wait(hdc, HDCS_DRDY, HDTIMEOUT_DRDY);
  if (result != 0) 
  {
    kprintf("hd_write_dma: no drdy (%d)\n", result);
    release_mutex(&hdc->lock);
    return -1;
  }

  // Setup DMA
  _outp((unsigned short) (hdc->bmregbase + BM_STATUS_REG), hdc->bmstat | BM_SR_MASK_INT | BM_SR_MASK_ERR);
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_STOP);
  setup_dma_transfer(hdc, buffer, count);
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_READ);

  // Prepare transfer
  result = 0;
  hdc->dir = HD_XFER_DMA;
  hdc->active = hd;

  hd_setup_transfer(hd, blkno, nsects);
  _outp((unsigned short) (hdc->iobase + HDC_COMMAND), HDCMD_WRITE_DMA);
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_READ | BM_CR_MASK_START);

  // Wait for interrupt
  wait_for_object(&hdc->ready);

  // Now wait for the PCI BM DMA channel to flush the data
  while (1)
  {
    if (_inp((unsigned short) (hdc->bmregbase + BM_STATUS_REG)) & BM_SR_MASK_INT) break;
  }

  // Check status
  hdc->status = _inp((unsigned short) (hdc->iobase + HDC_STATUS));
  if (hdc->status & HDCS_ERR)
  {
    kprintf("hd: dma write error (%d)\n", hdc->status);
    result = -1;
  }

  // Stop DMA channel
  _outp((unsigned short) (hdc->bmregbase + BM_COMMAND_REG), BM_CR_MASK_STOP);

  // Cleanup
  hdc->dir = HD_XFER_IDLE;
  hdc->active = NULL;
  release_mutex(&hdc->lock);
  return result == 0 ? count : -1;
}

void hd_dpc(void *arg)
{
  struct hdc *hdc = (struct hdc *) arg;

  switch (hdc->dir)
  {
    case HD_XFER_READ:
      // Read sector data
      pio_read_sect(hdc->active, hdc->bufp);

      // Check status
      hdc->status = _inp((unsigned short) (hdc->iobase + HDC_STATUS));
      if (hdc->status & HDCS_ERR)
      {
	kprintf("hd: read error (%d)\n", hdc->status);
	hdc->result = -1;
	set_event(&hdc->ready);
      }

      // Signal event if we have read all sectors
      hdc->bufp += SECTORSIZE;
      hdc->nsects--;
      if (hdc->nsects == 0) set_event(&hdc->ready);
      
      break;

    case HD_XFER_WRITE:
      // Check status
      hdc->status = _inp((unsigned short) (hdc->iobase + HDC_STATUS));
      if (hdc->status & HDCS_ERR)
      {
	kprintf("hd: write error (%d)\n", hdc->status);
	hdc->result = -1;
	set_event(&hdc->ready);
      }

      // Transfer next sector or signal end of transfer
      hdc->bufp += SECTORSIZE;
      hdc->nsects--;
      if (hdc->nsects > 0) 
	pio_write_sect(hdc->active, hdc->bufp);
      else
	set_event(&hdc->ready);

      break;

    case HD_XFER_DMA:
      set_event(&hdc->ready);
      break;

    case HD_XFER_IGNORE:
      break;

    case HD_XFER_IDLE:
    default:
      kprintf("unexpected intr from hdc\n");
  }
}

void hdc0_handler(struct context *ctxt)
{
  queue_irq_dpc(&hdctab[0].xfer_dpc, hd_dpc, &hdctab[0]);
  eoi(HDC0_IRQ);
}

void hdc1_handler(struct context *ctxt)
{
  queue_irq_dpc(&hdctab[1].xfer_dpc, hd_dpc, &hdctab[1]);
  eoi(HDC1_IRQ);
}

static int part_ioctl(struct dev *dev, int cmd, void *args)
{
  struct partition *part = (struct partition *) dev->privdata;

  switch (cmd)
  {
    case IOCTL_GETDEVSIZE:
      return part->len;

    case IOCTL_GETBLKSIZE:
      return dev_ioctl(part->dev, IOCTL_GETBLKSIZE, NULL);
  }

  return -1;
}

static int part_read(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct partition *part = (struct partition *) dev->privdata;
  if (blkno + count > part->len) return -1;
  return dev_read(part->dev, buffer, count, blkno + part->start);
}

static int part_write(struct dev *dev, void *buffer, size_t count, blkno_t blkno)
{
  struct partition *part = (struct partition *) dev->privdata;
  if (blkno + count > part->len) return -1;
  return dev_write(part->dev, buffer, count, blkno + part->start);
}

struct driver harddisk_dma_driver =
{
  "Harddisk (DMA)",
  DEV_TYPE_BLOCK,
  hd_ioctl,
  hd_read_dma,
  hd_write_dma
};

struct driver harddisk_intr_driver =
{
  "Harddisk", 
  DEV_TYPE_BLOCK,
  hd_ioctl,
  hd_read_intr,
  hd_write_intr
};

struct driver harddisk_poll_driver =
{
  "Harddisk (polling)", 
  DEV_TYPE_BLOCK,
  hd_ioctl,
  hd_read_poll,
  hd_write_poll
};

struct driver partition_driver =
{
  "Harddisk partition", 
  DEV_TYPE_BLOCK,
  part_ioctl,
  part_read,
  part_write
};

static void setup_hdc(struct hdc *hdc, int iobase, int irq, int bmregbase, void (*handler)(struct context *ctxt))
{
  memset(hdc, 0, sizeof(struct hdc));
  hdc->iobase = iobase;
  hdc->irq = irq;
  hdc->bmregbase = bmregbase;

  if (hdc->bmregbase)
  {
    // Save upper three bits of bus master status
    hdc->bmstat = _inp((unsigned short) (hdc->bmregbase + BM_STATUS_REG)) & 0xE0;

    // Allocate one page for PRD list
    hdc->prds = (struct prd *) kmalloc(PAGESIZE);
    hdc->prds_phys = (unsigned long) virt2phys(hdc->prds);
  }

  init_dpc(&hdc->xfer_dpc);
  init_mutex(&hdc->lock, 0);
  init_event(&hdc->ready, 0, 0);

  set_interrupt_handler(IRQ2INTR(irq), handler);
  enable_irq(irq);
  _outp((unsigned short) (iobase + HDC_DEVCTRL), HDDC_HD15);
}

static void setup_hd(struct hd *hd, struct hdc *hdc, char *devname, int drvsel, char *partnames[HD_PARTITIONS])
{
  int poll;
  int i;
  struct master_boot_record mbr;
  devno_t devno;
  devno_t partdevno;

  memset(hd, 0, sizeof(struct hd));
  hd->hdc = hdc;
  hd->drvsel = drvsel;

  hd_identify(hd);

  // Bochs emulator returns zero as vendor unique
  poll = 0;
  if (hd->param.vendorunique[0] == 0 &&
      hd->param.vendorunique[1] == 0 &&
      hd->param.vendorunique[2] == 0) poll = 1;

  if (poll)
    devno = dev_make(devname, &harddisk_poll_driver, hd);
  else if (hdc->bmregbase)
    devno = dev_make(devname, &harddisk_dma_driver, hd);
  else
    devno = dev_make(devname, &harddisk_intr_driver, hd);

  kprintf("%s: %u blks (%d MB) CHS=%u/%u/%u (%d bits, %s, %d sector buffers)\n", devname, hd->blks, hd->size, hd->cyls, hd->heads, hd->sectors, hd->use32bits ? 32 : 16, poll ? "poll" : (hdc->bmregbase ? "dma" : "intr"), hd->param.buffersize);

  hd_read_mbr(hd, &mbr);

  if (mbr.signature != MBR_SIGNATURE)
  {
    kprintf("%s: illegal boot sector signature\n", devname);
  }
  else
  {
    for (i = 0; i < HD_PARTITIONS; i++)
    {
      if (mbr.parts[i].systid != 0)
      {
	hd->parts[i].dev = devno;
	hd->parts[i].bootid = mbr.parts[i].bootid;
	hd->parts[i].systid = mbr.parts[i].systid;
	hd->parts[i].start = mbr.parts[i].relsect;
	hd->parts[i].len = mbr.parts[i].numsect;

	partdevno = dev_make(partnames[i], &partition_driver, &hd->parts[i]);
	if (bootdev == NODEV && mbr.parts[i].systid == 6) bootdev = partdevno;
 
	kprintf("%s: partition %d on %s, %dMB (type %02x)\n", partnames[i], i, devname, mbr.parts[i].numsect / (M / SECTORSIZE), mbr.parts[i].systid);
      }
    }

    if (bootdev == NODEV) bootdev = devno;
  }

  //kprintf("vendor: %02X %02X %02X model: [%s] rev: [%s] serial: [%s]\n", 
  //  hd->param.vendorunique[0], hd->param.vendorunique[1], hd->param.vendorunique[2],
  //  hd->param.model, hd->param.rev, hd->param.serial);
}

void init_hd()
{
  int bmiba;

  idedev = lookup_pci_device_class(PCI_CLASS_STORAGE_IDE, PCI_SUBCLASS_MASK);
  if (idedev)
  {
    bmiba = pci_config_read(idedev->bus->busno, idedev->devno, idedev->funcno, PCI_CONFIG_BASE_ADDR_4) & 0xFFF0;
  }

  setup_hdc(&hdctab[0], HDC0_IOBASE, HDC0_IRQ, idedev ? bmiba : 0, hdc0_handler);
  //setup_hdc(&hdctab[1], HDC1_IOBASE, HDC1_IRQ, idedev ? bmiba + 8: 0, hdc1_handler);

  setup_hd(&hdtab[0], &hdctab[0], "hd0", HD0_DRVSEL, hd0parts);
  //setup_hd(&hdtab[1], &hdctab[0], "hd1", HD1_DRVSEL, hd1parts);
  //setup_hd(&hdtab[2], &hdctab[1], "hd2", HD0_DRVSEL, hd2parts);
  //setup_hd(&hdtab[3], &hdctab[1], "hd3", HD1_DRVSEL, hd3parts);
}
