//
// os.h
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// Operating system API
//

#ifndef OS_H
#define OS_H

#include <types.h>
#include <stdarg.h>

#ifdef OS_LIB
#define apicall __declspec(dllexport)
#else
#ifdef KERNEL
#define apicall
#else
#define apicall __declspec(dllimport)
#endif
#endif

//
// Module initialization
//

#define DLL_PROCESS_ATTACH      1
#define DLL_THREAD_ATTACH       2
#define DLL_THREAD_DETACH       3
#define DLL_PROCESS_DETACH      0

//
// Page protection
//

#define PAGESIZE  4096

#define PAGE_NOACCESS           0x01     
#define PAGE_READONLY           0x02     
#define PAGE_READWRITE          0x04     
#define PAGE_EXECUTE            0x10     
#define PAGE_EXECUTE_READ       0x20     
#define PAGE_EXECUTE_READWRITE  0x40     
#define PAGE_GUARD              0x100     

//
// Allocation types
//

#define MEM_COMMIT              0x1000     
#define MEM_RESERVE             0x2000     
#define MEM_DECOMMIT            0x4000     
#define MEM_RELEASE             0x8000     

//
// File open modes
//

#define O_RDONLY                0x0000  // Open for reading only
#define O_WRONLY                0x0001  // Open for writing only
#define O_RDWR                  0x0002  // Open for reading and writing
#define O_APPEND                0x0008  // Writes done at EOF

#define O_CREAT                 0x0100  // Create and open file
#define O_TRUNC                 0x0200  // Truncate file
#define O_EXCL                  0x0400  // Open only if file doesn't already exist
#define O_DIRECT                0x0800  // Do not use cache for reads and writes

//
// Seek types
//

#define SEEK_SET                0       // Seek relative to begining of file
#define SEEK_CUR                1       // Seek relative to current file position
#define SEEK_END                2       // Seek relative to end of file

//
// Logging
//

#define LOG_LEVEL_MASK          0x0000000F
#define LOG_SUBSYS_MASK         0xFFFFFFF0

#define LOG_ERROR               1
#define LOG_WARNING             2
#define LOG_INFO                3
#define LOG_DEBUG               4
#define LOG_TRACE               5

#define LOG_MODULE              0x80000000
#define LOG_HEAP                0x40000000
#define LOG_APITRACE            0x20000000
#define LOG_AUX                 0x10000000

#define OK    0
#define INFINITE  0xFFFFFFFF

//
// Error codes (from MSVCRT)
//

//#define EPERM           1                // Operation not permitted
#define ENOENT          2                // No such file or directory
//#define ESRCH           3                // No such process
//#define EINTR           4                // Interrupted system call
//#define EIO             5                // Input/output error
//#define ENXIO           6                // Device not configured
//#define E2BIG           7                // Argument list too long
//#define ENOEXEC         8                // Exec format error
#define EBADF           9                // Bad file number
//#define ECHILD          10               // No spawned processes
//#define EAGAIN          11               // Resource temporarily unavailable
#define ENOMEM          12               // Cannot allocate memory
#define EACCES          13               // Access denied
#define EFAULT          14               // Bad address
//#define E0015           15               // Error 15
//#define EBUSY           16               // Device busy
#define EEXIST          17               // File exist
#define EXDEV           18               // Cross-device link
#define ENODEV          19               // Operation not supported by device
//#define ENOTDIR         20               // Not a directory
//#define EISDIR          21               // Is a directory
#define EINVAL          22               // Invalid argument
#define ENFILE          23               // Too many open files in system
//#define EMFILE          24               // Too many files open
//#define ENOTTY          25               // Inappropriate ioctl for device
//#define E0026           26               // Error 26
//#define EFBIG           27               // File too large
//#define ENOSPC          28               // No space left on device
//#define ESPIPE          29               // Illegal seek
//#define EROFS           30               // Read-only file system
//#define EMLINK          31               // Too many links
//#define EPIPE           32               // Broken pipe
//#define EDOM            33               // Numerical arg out of domain
//#define ERANGE          34               // Result too large
//#define E0035           35               // Error 35
//#define EDEADLK         36               // Resource deadlock avoided
//#define E0037           37               // Error 37
#define ENAMETOOLONG    38               // File name too long
//#define ENOLCK          39               // No locks available
#define ENOSYS          40               // Function not implemented
//#define ENOTEMPTY       41               // Directory not empty
//#define EILSEQ          42               // Invalid multibyte sequence

//
// File status
//

#define MAXPATH                 256     // Maximum filename length (including trailing zero)

#define PS1                     '/'    // Primary path separator
#define PS2                     '\\'   // Alternate path separator

struct stat
{
  int mode;
  ino_t ino;
  time_t atime;
  time_t mtime;
  time_t ctime;
  union
  {
    struct
    {
      unsigned long size_low;
      unsigned long size_high;
    } quad;
    unsigned __int64 size;
  };
};

//
// Directory entry
//

struct dirent
{
  ino_t ino;
  unsigned int reclen;
  unsigned int namelen;
  char name[MAXPATH];
};

//
// Critical sections
//

struct critsect
{
  long count;
  long recursion;
  unsigned long owner;
  handle_t event;
};

typedef struct critsect *critsect_t;

// Thread Information Block

#define MAX_TLS 64
#define INVALID_TLS_INDEX 0xFFFFFFFF

struct xcptrec
{
  struct xcptrec *next;
  void (*handler)();
};

struct tib
{
  struct xcptrec *except;
  void *stacktop;
  void *stackbase;

  void *subsystemtib;
  unsigned long fiberdata;

  void *arbitrary;
  struct tib *self;

  unsigned long unknown1;
  unsigned long pid;
  unsigned long tid;
  unsigned long unknown2;

  void *tlsbase;

  void *stackend;
  char reserved1[3548];

  void *tls[MAX_TLS];
  char reserved2[240];
};

#ifndef KERNEL

// OS API functions

apicall int syscall(int syscallno, void *params);

apicall int format(char *devname, char *type, char *opts);
apicall int mount(char *type, char *path, char *devname, char *opts);
apicall int unmount();

apicall handle_t open(char *name, int mode);
apicall int close(handle_t h);
apicall int flush(handle_t f);

apicall int read(handle_t f, void *data, size_t size);
apicall int write(handle_t f, void *data, size_t size);

apicall loff_t tell(handle_t f);
apicall loff_t lseek(handle_t f, loff_t offset, int origin);
apicall int chsize(handle_t f, loff_t size);

apicall int fstat(handle_t f, struct stat *buffer);
apicall int stat(char *name, struct stat *buffer);

apicall int mkdir(char *name);
apicall int rmdir(char *name);

apicall int rename(char *oldname, char *newname);
apicall int link(char *oldname, char *newname);
apicall int unlink(char *name);

apicall handle_t opendir(char *name);
apicall int readdir(handle_t f, struct dirent *dirp, int count);

apicall void *mmap(void *addr, unsigned long size, int type, int protect);
apicall int munmap(void *addr, unsigned long size, int type);
apicall void *mremap(void *addr, unsigned long oldsize, unsigned long newsize, int type, int protect);
apicall int mprotect(void *addr, unsigned long size, int protect);
apicall int mlock(void *addr, unsigned long size);
apicall int munlock(void *addr, unsigned long size);

apicall int wait(handle_t h, int timeout);

apicall handle_t mkevent(int manual_reset, int initial_state);
apicall int epulse(handle_t h);
apicall int eset(handle_t h);
apicall int ereset(handle_t h);

apicall handle_t self();
apicall void exit();

apicall void panic(char *msg);
apicall void syslog(int priority, const char *fmt,...);
apicall char *canonicalize(char *filename, char *buffer, int size);

apicall void mkcs(critsect_t cs);
apicall void csfree(critsect_t cs);
apicall void enter(critsect_t cs);
apicall void leave(critsect_t cs);

apicall void *malloc(size_t size);
apicall void *realloc(void *mem, size_t size);
apicall void *calloc(size_t num, size_t size);
apicall void free(void *p);

apicall void *resolve(hmodule_t hmod, char *procname);
apicall hmodule_t getmodule(char *name);
apicall int getmodpath(hmodule_t hmod, char *buffer, int size);
apicall hmodule_t load(char *name);

apicall tls_t tlsalloc();
apicall void tlsfree(tls_t index);
apicall void *tlsget(tls_t index);
apicall void tlsset(tls_t index, void *value);

#endif

#endif
