//
// vfs.h
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// Virtual filesystem
//

#ifndef VFS_H
#define VFS_H

#define F_MODIFIED              0x1000  // File has been modified since it was opened
#define F_DIR                   0x2000  // File is a directory

#define FS_DIRECTORY            1       // File is a directory
#define FS_BLKDEV               2       // File is a block device
#define FS_STREAMDEV            4       // File is a stream device

struct vfsops;

struct filesystem
{
  char *name;
  struct fsops *ops;
  struct filesystem *next;
};

struct fs
{
  devno_t devno;
  char path[MAXPATH + 1];
  struct fsops *ops;
  struct fs *next;
  void *data;
};

struct file
{
  struct object object;

  struct fs *fs;
  int flags;
  loff_t pos;
  void *data;
};

struct fsops
{
  int (*format)(devno_t devno, char *opts);
  int (*mount)(struct fs *fs, char *opts);
  int (*unmount)(struct fs *fs);

  int (*open)(struct file *filp, char *name);
  int (*close)(struct file *filp);
  int (*flush)(struct file *filp);

  int (*read)(struct file *filp, void *data, size_t size);
  int (*write)(struct file *filp, void *data, size_t size);

  loff_t (*tell)(struct file *filp);
  loff_t (*lseek)(struct file *filp, loff_t offset, int origin);
  int (*chsize)(struct file *filp, loff_t size);

  int (*fstat)(struct file *filp, struct stat *buffer);
  int (*stat)(struct fs *fs, char *name, struct stat *buffer);

  int (*mkdir)(struct fs *fs, char *name);
  int (*rmdir)(struct fs *fs, char *name);

  int (*rename)(struct fs *fs, char *oldname, char *newname);
  int (*link)(struct fs *fs, char *oldname, char *newname);
  int (*unlink)(struct fs *fs, char *name);
  
  int (*opendir)(struct file *filp, char *name);
  int (*readdir)(struct file *filp, struct dirent *dirp, int count);
};

#ifdef KERNEL

int fnmatch(char *fn1, int len1, char *fn2, int len2);
struct filesystem *register_filesystem(char *name, struct fsops *ops);
struct fs *fslookup(char *name, char **rest);

int format(devno_t devno, char *type, char *opts);
int mount(char *type, char *path, devno_t devno, char *opts);
int unmount_all();

int open(char *name, int mode, struct file **retval);
int close(struct file *filp);
int flush(struct file *filp);

int read(struct file *filp, void *data, size_t size);
int write(struct file *filp, void *data, size_t size);

loff_t tell(struct file *filp);
loff_t lseek(struct file *filp, loff_t offset, int origin);
int chsize(struct file *filp, loff_t size);

int fstat(struct file *filp, struct stat *buffer);
int stat(char *name, struct stat *buffer);

int mkdir(char *name);
int rmdir(char *name);

int rename(char *oldname, char *newname);
int link(char *oldname, char *newname);
int unlink(char *name);

int opendir(char *name, struct file **retval);
int readdir(struct file *filp, struct dirent *dirp, int count);

#endif

#endif
