//
// pci.c
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// PCI bus interface
//

#include <os/krnl.h>

#define MAX_PCI_BUSSES 8

#define PCI_BRIDGE 0x60400

struct pci_bus *pci_root;
struct pci_bus *pci_busses[MAX_PCI_BUSSES];
int num_pci_busses;

unsigned long pci_config_read(int busno, int devno, int funcno, int addr)
{
  _outpd(PCI_CONFIG_ADDR, ((unsigned long) 0x80000000 | (busno << 16) | (devno << 11) | (funcno << 8) | (addr << 2)));
  return _inpd(PCI_CONFIG_DATA);
}

void pci_config_write(int busno, int devno, int funcno, int addr, unsigned long value)
{
  _outpd(PCI_CONFIG_ADDR, ((unsigned long) 0x80000000 | (busno << 16) | (devno << 11) | (funcno << 8) | (addr << 2)));
  _outpd(PCI_CONFIG_DATA, value);
}

struct pci_dev *lookup_pci_device(unsigned short vendorid, unsigned short deviceid)
{
  int i;
  struct pci_bus *bus;
  struct pci_dev *dev;

  for (i = 0; i < num_pci_busses; i++)
  {
    bus = pci_busses[i];
    dev = bus->devices;

    while (dev)
    {
      if (dev->vendorid == vendorid && dev->deviceid == deviceid) return dev;
      dev = dev->next;
    }

  }

  return NULL;
}

struct pci_dev *lookup_pci_device_class(int classcode, int mask)
{
  int i;
  struct pci_bus *bus;
  struct pci_dev *dev;

  for (i = 0; i < num_pci_busses; i++)
  {
    bus = pci_busses[i];
    dev = bus->devices;

    while (dev)
    {
      if ((dev->classcode & mask) == classcode) return dev;
      dev = dev->next;
    }

  }

  return NULL;
}

static void scan_pci_bus(struct pci_bus *bus)
{
  int devno;
  int funcno;
  int bar;
  unsigned long value;
  unsigned short vendorid;
  unsigned short deviceid;
  struct pci_dev *dev;
  struct pci_dev *devtail;
  struct pci_bus *bustail;

  devtail = NULL;
  bustail = NULL;

  for (devno = 0; devno < 32; devno++)
  {
    for (funcno = 0; funcno < 8; funcno++)
    {
      // Vendor and device ids
      value = pci_config_read(bus->busno, devno, funcno, PCI_CONFIG_VENDOR);
      vendorid = (unsigned short) (value & 0xFFFF);
      deviceid = (unsigned short) (value >> 16);

      if (vendorid != 0xFFFF && vendorid != 0)
      {
	if (funcno == 0 || devtail == NULL || devtail->devno != devno || devtail->deviceid != deviceid)
	{
	  // Allocate and initialize new PCI device
	  dev = (struct pci_dev *) kmalloc(sizeof(struct pci_dev));
	  memset(dev, 0, sizeof(struct pci_dev));

	  dev->bus = bus;
	  dev->devno = devno;
	  dev->funcno = funcno;
	  dev->vendorid = vendorid;
	  dev->deviceid = deviceid;

	  if (devtail) 
	    devtail->next = dev;
	  else
	    bus->devices = dev;
	  devtail = dev;

	  // Function class code
	  value = pci_config_read(bus->busno, devno, funcno, PCI_CONFIG_CLASS_REV);
	  dev->classcode = value >> 8;

	  if (dev->classcode == PCI_BRIDGE)
	  {
	    struct pci_bus *bridge;

	    // PCI-PCI bridge
	    if (num_pci_busses == MAX_PCI_BUSSES) panic("too many pci busses");

            // Allocate and initialize new PCI bus
	    bridge = (struct pci_bus *) kmalloc(sizeof(struct pci_bus));
            memset(bridge, 0, sizeof(struct pci_bus));
	    pci_busses[num_pci_busses++] = bridge;

	    bridge->parent = bus;
	    bridge->self = dev;

	    if (bustail)
	      bustail->next = bridge;
	    else
	      bus->bridges = bridge;
	    bustail = bridge;

	    // Get secondary bus number for bridge
            value = pci_config_read(bus->busno, devno, funcno, PCI_CONFIG_BASE_ADDR_2);
	    bridge->busno = (value >> 8) & 0xFF;

	    // Scan for devices on secondary bus
	    scan_pci_bus(bridge);
	  }
	  else
	  {
	    // Function i/o and memory base addresses
	    for (bar = 0; bar < 6; bar++)
	    {
  	      value = pci_config_read(bus->busno, devno, funcno, PCI_CONFIG_BASE_ADDR_0 + bar);
	      if (value != 0)
	      {
		if (value & 1)
		{
		  if (dev->iobase == 0) dev->iobase = value & 0xFFFFFFFC;
		}
		else
		{
		  if (dev->membase == 0) dev->membase = value & 0xFFFFFFFC;
		}
	      }

	    }

	    // Function interrupt line
	    value = pci_config_read(bus->busno, devno, funcno, PCI_CONFIG_INTR);
	    if ((value & 0xFF) > 0 && (value & 0xFF) < 32)
	    {
	      dev->intrpin = (value >> 8) & 0xFF;
	      dev->irq = value & 0xFF;
	    }
	  }
	}
      }
    }
  }
}

void init_pci()
{
  pci_root = (struct pci_bus *) kmalloc(sizeof(struct pci_bus));
  memset(pci_root, 0, sizeof(struct pci_bus));
 
  pci_busses[0] = pci_root;
  num_pci_busses = 1;

  scan_pci_bus(pci_root);
}

void dump_pci_devices()
{
  int i;
  struct pci_bus *bus;
  struct pci_dev *dev;

  for (i = 0; i < num_pci_busses; i++)
  {
    bus = pci_busses[i];
    dev = bus->devices;

    while (dev)
    {
      kprintf("dev %d.%d.%d vendor %X device %X class %X", dev->bus->busno, dev->devno, dev->funcno, dev->vendorid, dev->deviceid, dev->classcode);

      if (dev->classcode == PCI_BRIDGE)
      {
	kprintf(" bridge\n");
      }
      else
      {
	if (dev->membase != 0) kprintf(" mem %X", dev->membase);
	if (dev->iobase != 0) kprintf(" io %X", dev->iobase);
	if (dev->irq != 0) kprintf(" irq %d", dev->irq);
	kprintf("\n");
      }

      dev = dev->next;
    }
  }
}
