//
// sched.c
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// Task scheduler
//

#include <os/krnl.h>

int resched = 0;
int idle = 0;
struct thread *idle_thread;
struct thread *ready_queue_head[THREAD_PRIORITY_LEVELS];
struct thread *ready_queue_tail[THREAD_PRIORITY_LEVELS];
struct dpc *dpc_queue;

__declspec(naked) void context_switch(struct thread *t)
{
  __asm
  {
    // Save register on current kernel stack
    push    ebp
    push    ebx
    push    edi
    push    esi

    // Store kernel stack pointer in tcb
    mov	    eax, esp
    and	    eax, TCBMASK
    add	    eax, TCBESP
    mov	    [eax], esp

    // Get stack pointer for new thread and store in esp0
    mov	    eax, 20[esp]
    add	    eax, TCBESP
    mov	    esp, [eax]
    mov	    ebp, TSS_ESP0
    mov	    [ebp], eax

    // Restore register from new kernel stack
    pop	    esi
    pop	    edi
    pop	    ebx
    pop	    ebp

    ret
  }
}

__declspec(naked) struct thread *current_thread()
{
  __asm
  {
    mov eax, esp;
    and eax, TCBMASK
    ret
  }
}

static void init_thread_stack(struct thread *t, void *startaddr, void *arg)
{
  struct tcb *tcb = (struct tcb *) t;
  unsigned long *esp = (unsigned long *) &tcb->esp;
  
  *--esp = (unsigned long) arg;
  *--esp = (unsigned long) 0;
  *--esp = (unsigned long) startaddr;
  *--esp = 0;
  *--esp = 0;
  *--esp = 0;
  *--esp = 0;
  tcb->esp = esp;
}

void mark_thread_ready(struct thread *t)
{
  t->state = THREAD_STATE_READY;
  t->next_ready = NULL;

  if (ready_queue_tail[t->priority] != NULL) ready_queue_tail[t->priority]->next_ready = t;
  if (ready_queue_head[t->priority] == NULL) ready_queue_head[t->priority] = t;
  ready_queue_tail[t->priority] = t;
  resched = 1;
}

void mark_thread_running()
{
  struct thread *self = current_thread();
  struct tib *tib = self->tib;
  struct segment *seg;

  // Set thread state to running
  self->state = THREAD_STATE_RUNNING;

  // Set FS register to point to current TIB
  seg = &syspage->gdt[GDT_TIB];
  seg->base_low = (unsigned short)((unsigned long) tib & 0xFFFF);
  seg->base_med = (unsigned char)(((unsigned long) tib >> 16) & 0xFF);
  seg->base_high = (unsigned char)(((unsigned long) tib >> 24) & 0xFF);

  // Reload FS register
  __asm
  {
    mov ax, SEL_TIB + SEL_RPL3
    mov fs, ax
  }
}

struct thread *create_task(taskproc_t task, void *arg, int priority)
{
  // Allocate a new aligned thread control block
  struct thread *t = (struct thread *) alloc_pages_align(PAGES_PER_TCB, PAGES_PER_TCB);
  memset(t, 0, PAGES_PER_TCB * PAGESIZE);
  init_thread(t, priority);

  // Initialize the thread kernel stack to start executing the task function
  init_thread_stack(t, (void *) task, arg);

  // Mark new thread as ready to run
  mark_thread_ready(t);
  
  return t;
}

int init_user_thread(struct thread *t, unsigned long stack_reserved, unsigned long stack_committed)
{
  char *stack;
  handle_t h;
  struct tib *tib;

  // Allocate and initialize thread information block for thread
  tib = mmap(NULL, sizeof(struct tib), MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
  if (!tib) return -ENOMEM;

  t->tib = tib;
  tib->self = tib;
  tib->tlsbase = &tib->tls;
  tib->pid = 1;
  tib->tid = t->id;

  // Allocate thread handle
  h = halloc(&t->object);
  if (h < 0)  return h;
  t->self = h;

  // Allocate user stack for thread
  stack = mmap(NULL, stack_reserved, MEM_RESERVE, PAGE_READWRITE);
  if (!stack) return -ENOMEM;

  tib->stackend = stack;
  tib->stacktop = stack + stack_reserved;
  tib->stackbase = stack + (stack_reserved - stack_committed);

  if (!mmap(tib->stackbase, stack_committed, MEM_COMMIT, PAGE_READWRITE)) return -ENOMEM;

  return 0;
}

void init_dpc(struct dpc *dpc)
{
  dpc->proc = NULL;
  dpc->arg = NULL;
  dpc->next = NULL;
}

void queue_dpc(struct dpc *dpc, dpcproc_t proc, void *arg)
{
  __asm { cli };
  dpc->proc = proc;
  dpc->arg = arg;
  dpc->next = dpc_queue;
  dpc_queue = dpc;
  __asm { sti };
  resched = 1;
}

void queue_irq_dpc(struct dpc *dpc, dpcproc_t proc, void *arg)
{
  dpc->proc = proc;
  dpc->arg = arg;
  dpc->next = dpc_queue;
  dpc_queue = dpc;
  resched = 1;
}

void dispatch_dpc_queue()
{
  struct dpc *queue;
  struct dpc *next;

  // Get list of queued deferred procedure calls
  __asm { cli };
  queue = dpc_queue;
  dpc_queue = NULL;
  __asm { sti };

  // Execute each DPC
  while (queue)
  {
    next = queue->next;
    queue->proc(queue->arg);
    queue = next;
  }
}

void dispatch()
{
  int prio;
  struct thread *t;

  // Clear rescheduling flag
  resched = 0;

  // Execute all queued DPC's
  dispatch_dpc_queue();

  // Find next thread to run
  prio = THREAD_PRIORITY_LEVELS - 1;
  while (ready_queue_head[prio] == 0 && prio > 0) prio--;
  t = ready_queue_head[prio];
  if (t == NULL) panic("No thread ready to run");

  // Remove thread from ready queue
  ready_queue_head[prio] = t->next_ready;
  if (t->next_ready == NULL) ready_queue_tail[prio] = NULL;
  t->next_ready = NULL;

  // If current thread has been selected to run again then just return
  if (t == current_thread()) return;

  // Switch to new thread
  context_switch(t);

  // Mark new thread as running
  mark_thread_running();
}

void yield()
{
  // Mark thread as ready to run
  mark_thread_ready(current_thread());

  // Dispatch next thread
  dispatch();
}

void idle_task()
{
  while (1) 
  {
    idle = 1;
    __asm { hlt };
    idle = 0;
    if (resched)
    {
      mark_thread_ready(current_thread());
      dispatch();
    }
  }
}

void debug_intr_handler(struct context *ctxt)
{
  //kprintf("debug intr\n");
}

void init_sched()
{
  // Initialize scheduler
  resched = 0;
  dpc_queue = NULL;
  memset(ready_queue_head, 0, sizeof(ready_queue_head));
  memset(ready_queue_tail, 0, sizeof(ready_queue_tail));

  // The initial kernel thread will later become the idle thread
  idle_thread = current_thread();

  // The idle thread is always ready to run
  memset(idle_thread, 0, sizeof(struct thread));
  idle_thread->object.type = OBJECT_THREAD;
  idle_thread->priority = PRIORITY_IDLE;
  idle_thread->state = THREAD_STATE_RUNNING;

  set_interrupt_handler(INTR_DEBUG, debug_intr_handler);
}
