//
// start.c
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// Kernel initialization
//

#include <os/krnl.h>

#define VERSION "0.0.2"

#define INIT_STACK_COMMITTED (64 * K)
#define INIT_STACK_RESERVED  (64 * K)

struct thread *mainthread;
devno_t bootdev = NODEV;

struct file *stdin;
struct file *stdout;
struct file *stderr;

#define RVA(hmod, rva) (((char *) (hmod)) + (rva))

void shell();

void main(void *arg);

void panic(char *msg)
{
  kprintf("panic: %s\n", msg);
  __asm { cli };
  __asm { hlt };
}

void exit()
{
  kprintf("syncing filesystems...\n");
  unmount_all();
  kprintf("system stopped\n");
  sleep(100);
  __asm { cli };
  __asm { hlt };
}

static void *load_image(char *filename, void **entrypoint)
{
  struct file *f;
  char *buffer;
  char *imgbase;
  struct dos_header *doshdr;
  struct image_header *imghdr;
  int i;

  // Allocate header buffer
  buffer = kmalloc(PAGESIZE);
  if (!buffer) return NULL;

  // Open file
  if (open(filename, O_RDONLY, &f) < 0) 
  {
    kfree(buffer);
    return NULL;
  }

  // Read headers
  if (read(f, buffer, PAGESIZE) < 0)
  {
    close(f);
    kfree(buffer);
    return NULL;
  }
  doshdr = (struct dos_header *) buffer;
  imghdr = (struct image_header *) (buffer + doshdr->e_lfanew);

  // Check PE file signature
  if (imghdr->signature != IMAGE_PE_SIGNATURE) panic("invalid PE signature");

  // Check alignment
  if (imghdr->optional.file_alignment != PAGESIZE || imghdr->optional.section_alignment != PAGESIZE) panic("image not page aligned");

kprintf("load address is %p, size is %d\n", imghdr->optional.image_base, imghdr->optional.size_of_image);

  // Allocate memory for module
  imgbase = (char *) mmap((void *) (imghdr->optional.image_base), imghdr->optional.size_of_image, MEM_RESERVE | MEM_COMMIT, PAGE_EXECUTE_READWRITE);
kprintf("imgbase is %p\n", imgbase);
  if (imgbase == NULL)
  {
    close(f);
    kfree(buffer);
    return NULL;
  }

  // copy header to image
  memcpy(imgbase, buffer, PAGESIZE);

  // Read sections
  for (i = 0; i < imghdr->header.number_of_sections; i++)
  {
    int protect;
    unsigned long  scn = imghdr->sections[i].characteristics & (IMAGE_SCN_MEM_EXECUTE | IMAGE_SCN_MEM_READ | IMAGE_SCN_MEM_WRITE);

    if (scn == (IMAGE_SCN_MEM_EXECUTE | IMAGE_SCN_MEM_READ)) 
      protect = PAGE_EXECUTE_READ;
    else if (scn == (IMAGE_SCN_MEM_READ | IMAGE_SCN_MEM_WRITE)) 
      protect = PAGE_READWRITE;
    else if (scn == IMAGE_SCN_MEM_READ)
      protect = PAGE_READONLY;
    else
      protect = PAGE_EXECUTE_READWRITE;

kprintf("read section %d: %s ", i, imghdr->sections[i].name);
if (imghdr->sections[i].characteristics & IMAGE_SCN_MEM_EXECUTE) kprintf("execute ");
if (imghdr->sections[i].characteristics & IMAGE_SCN_MEM_READ) kprintf("read ");
if (imghdr->sections[i].characteristics & IMAGE_SCN_MEM_WRITE) kprintf("write ");
kprintf("\n");

    if (imghdr->sections[i].pointer_to_raw_data != 0)
    {
      lseek(f, imghdr->sections[i].pointer_to_raw_data, SEEK_SET);
      if (read(f, RVA(imgbase, imghdr->sections[i].virtual_address), imghdr->sections[i].size_of_raw_data) < 0)
      {
	munmap(imgbase, imghdr->optional.size_of_image, MEM_RELEASE);
	close(f);
        kfree(buffer);
	return NULL;
      }
    }

    mprotect(RVA(imgbase, imghdr->sections[i].virtual_address), imghdr->sections[i].size_of_raw_data, protect); 
  }

  // Find entry point
  *entrypoint = RVA(imgbase, imghdr->optional.address_of_entry_point);

kprintf("image %s loaded at %p (%d KB) entrypoint at %p\n", filename, imgbase, imghdr->optional.size_of_image / K, *entrypoint);
  
  // Close file
  close(f);
  kfree(buffer);

  return imgbase;
}

void __stdcall start(void *hmod, int memend, int heap)
{
  init_video();
  clear_screen();

  print_string("SanOS version " VERSION " (Build " __DATE__ " " __TIME__ ")\n");
  print_string("Copyright (c) 2001 Michael Ringgaard. All rights reserved.\n\n");

  //kprintf("Initializing page frame database\n");
  init_pfdb(heap, memend);

  //kprintf("Initializing page directory\n");
  init_pdir();

  //kprintf("Initializing kernel heap\n");
  init_kmem();

  //kprintf("Initializing kernel allocator\n");
  init_malloc();

  //kprintf("Initializing virtual memory manager\n");
  init_vmm();

  // flush tlb
  flushtlb();

  //kprintf("Initializing devices\n");
  init_pic();
  init_intr();
  init_pit();
  init_keyboard();
  
  //kprintf("Initializing scheduler\n");
  init_sched();

  // Enable interrupts
  __asm { sti };

  init_pci();
  init_cons();
  init_null();
  init_hd();
  init_dfs();
  init_devfs();

  kprintf("Memory %dMB total, %dKB used, %dKB free, %dKB reserved\n", 
	  maxmem * PAGESIZE / M, 
	  (totalmem - freemem) * PAGESIZE / K, 
	  freemem * PAGESIZE / K, (maxmem - totalmem) * PAGESIZE / K);

  // Start main task and dispatch to idle task
  mainthread = create_task(main, 0, PRIORITY_NORMAL);
  idle_task();
}

void main(void *arg)
{
  void *imgbase;
  void *entrypoint;
  unsigned long *stacktop;
  struct thread *self;

  struct tcb *tcb;

  // Check for boot device
  if (bootdev == NODEV) panic("no boot device found");

  // Mount file systems
  mount("dfs", "/", bootdev, "");
  mount("devfs", "/dev", NODEV, NULL);

  // Allocate handles for stdin, stdout and stderr
  open("/dev/console", O_RDONLY, &stdin);
  open("/dev/console", O_WRONLY, &stdout);
  open("/dev/console", O_WRONLY, &stderr);
  if (halloc(&stdin->object) != 0) panic("unexpected stdin handle");
  if (halloc(&stdout->object) != 1) panic("unexpected stdout handle");
  if (halloc(&stderr->object) != 2) panic("unexpected stderr handle");

  // Load os.dll in user address space
  imgbase = load_image("/os/os.dll", &entrypoint);
  if (!imgbase) panic("unable to load os.dll");

  // Initialize initial user thread
  self = current_thread();
  if (init_user_thread(self, INIT_STACK_RESERVED, INIT_STACK_COMMITTED) < 0)
  {
    panic("unable to initialize initial user thread");
  }
  mark_thread_running();

  // Place arguments to start routine on stack
  stacktop = (unsigned long *) self->tib->stacktop;
  *(--stacktop) = 0;
  *(--stacktop) = 0;
  *(--stacktop) = (unsigned long) imgbase;
  *(--stacktop) = 0;

tcb = (struct tcb *) self;
kprintf("self %p esp %p id %d stack %p stacktop %p\n", self, tcb->esp, self->id, self->tib->stackend, stacktop);

kprintf("sizeof(tib) = %d offset(self) = %x offset(tls) = %x\n", sizeof(struct tib), &((struct tib *) 0)->self, &((struct tib *) 0)->tls);

shell();
  // Jump into user mode
  __asm
  {
    mov eax, stacktop
    mov ebx, entrypoint

    push SEL_UDATA + SEL_RPL3
    push eax
    pushfd
    push SEL_UTEXT + SEL_RPL3
    push ebx
    iretd

    cli
    hlt
  }
}
