//
// syscall.c
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// System call interface
//

#include <os/krnl.h>

static int lock_buffer(void *buffer, int size)
{
  return 0;
}

static int unlock_buffer(void *buffer, int size)
{
  return 0;
}

static int lock_string(char *s)
{
  return 0;
}

static int unlock_string(char *s)
{
  return 0;
}

static int sys_format(char *params)
{
  char *devname; 
  char *type;
  char *opts;
  int rc;
  devno_t dev;

  if (lock_buffer(params, 12) < 0) return -EFAULT;

  devname = *(char **) params;
  type = *(char **) (params + 4);
  opts = *(char **) (params + 8);

  if (lock_string(devname) < 0)
  {
    unlock_buffer(params, 12);
    return -EFAULT;
  }

  if (lock_string(type) < 0)
  {
    unlock_string(devname);
    unlock_buffer(params, 12);
    return -EFAULT;
  }

  if (lock_string(opts) < 0)
  {
    unlock_string(type);
    unlock_string(devname);
    unlock_buffer(params, 12);
    return -EFAULT;
  }

  dev = dev_open(devname);
  if (dev != NODEV)
  {
    rc = format(dev, type, opts);
    dev_close(dev);
  }
  else
    rc = -ENOENT;

  unlock_string(opts);
  unlock_string(type);
  unlock_string(devname);
  unlock_buffer(params, 12);

  return rc;
}

static int sys_mount(char *params)
{
  char *type;
  char *path;
  char *devname; 
  char *opts;
  int rc;
  devno_t dev;

  if (lock_buffer(params, 16) < 0) return -EFAULT;

  type = *(char **) params;
  path = *(char **) (params + 4);
  devname = *(char **) (params + 8);
  opts = *(char **) (params + 12);

  if (lock_string(type) < 0)
  {
    unlock_buffer(params, 16);
    return -EFAULT;
  }

  if (lock_string(path) < 0)
  {
    unlock_string(type);
    unlock_buffer(params, 16);
    return -EFAULT;
  }

  if (lock_string(devname) < 0)
  {
    unlock_string(path);
    unlock_string(type);
    unlock_buffer(params, 16);
    return -EFAULT;
  }

  if (lock_string(opts) < 0)
  {
    unlock_string(devname);
    unlock_string(path);
    unlock_string(type);
    unlock_buffer(params, 16);
    return -EFAULT;
  }

  dev = dev_open(devname);
  if (dev != NODEV)
    rc = mount(type, path, dev, opts);
  else
    rc = -ENOENT;

  unlock_string(opts);
  unlock_string(devname);
  unlock_string(path);
  unlock_string(type);
  unlock_buffer(params, 16);

  return rc;
}

static int sys_unmount(char *params)
{
  return -ENOSYS;
}

static int sys_open(char *params)
{
  struct file *f;
  char *name;
  int mode;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  name = *(char **) params;
  mode = *(int *) (params + 4);

  if (lock_string(name) < 0)
  {
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  rc = open(name, mode, &f);
  if (rc == 0)
  {
    rc = halloc(&f->object);
    if (rc < 0) close(f);
  }

  unlock_string(name);
  unlock_buffer(params, 8);

  return rc;
}

static int sys_close(char *params)
{
  handle_t h;
  int rc;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  h = *(handle_t *) params;

  rc = hfree(h);

  unlock_buffer(params, 4);

  return rc;
}

static int sys_flush(char *params)
{
  struct file *f;
  handle_t h;
  int rc;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  h = *(handle_t *) params;
  f = (struct file *) hlock(h, OBJECT_FILE);
  if (!f) 
  {
    unlock_buffer(params, 4);
    return -EBADF;
  }

  rc = flush(f);

  hrel(h);
  unlock_buffer(params, 4);

  return rc;
}

static int sys_read(char *params)
{
  handle_t h;
  struct file *f;
  int rc;
  void *data;
  int size;

  if (lock_buffer(params, 12) < 0) return -EFAULT;

  h = *(handle_t *) params;
  data = *(void **) (params + 4);
  size = *(int *) (params + 8);

  f = (struct file *) hlock(h, OBJECT_FILE);
  if (!f) 
  {
    unlock_buffer(params, 12);
    return -EBADF;
  }

  if (lock_buffer(data, size) < 0)
  {
    hrel(h);
    unlock_buffer(params, 12);
    return -EFAULT;
  }
  
  rc = read(f, data, size);
  
  hrel(h);
  unlock_buffer(params, 12);

  return rc;
}

static int sys_write(char *params)
{
  handle_t h;
  struct file *f;
  int rc;
  void *data;
  int size;

  if (lock_buffer(params, 12) < 0) return -EFAULT;

  h = *(handle_t *) params;
  data = *(void **) (params + 4);
  size = *(int *) (params + 8);

  f = (struct file *) hlock(h, OBJECT_FILE);
  if (!f) 
  {
    unlock_buffer(params, 12);
    return -EBADF;
  }

  if (lock_buffer(data, size) < 0)
  {
    hrel(h);
    unlock_buffer(params, 12);
    return -EFAULT;
  }
  
  rc = write(f, data, size);
  
  hrel(h);
  unlock_buffer(params, 12);

  return rc;
}

static int sys_tell(char *params)
{
  struct file *f;
  handle_t h;
  loff_t retval;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  h = *(handle_t *) params;
  f = (struct file *) hlock(h, OBJECT_FILE);
  if (!f) 
  {
    unlock_buffer(params, 4);
    return -EBADF;
  }

  retval = flush(f);

  hrel(h);
  unlock_buffer(params, 4);

  return (int) retval;
}

static int sys_lseek(char *params)
{
  struct file *f;
  handle_t h;
  loff_t offset;
  int origin;
  loff_t retval;

  if (lock_buffer(params, 12) < 0) return -EFAULT;

  h = *(handle_t *) params;
  offset = *(loff_t *) (params + 4);
  origin = *(int *) (params + 8);

  f = (struct file *) hlock(h, OBJECT_FILE);
  if (!f) 
  {
    unlock_buffer(params, 12);
    return -EBADF;
  }

  retval = lseek(f, offset, origin);

  hrel(h);
  unlock_buffer(params, 12);

  return (int) retval;
}

static int sys_chsize(char *params)
{
  struct file *f;
  handle_t h;
  loff_t size;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  h = *(handle_t *) params;
  size = *(loff_t *) (params + 4);

  f = (struct file *) hlock(h, OBJECT_FILE);
  if (!f) 
  {
    unlock_buffer(params, 8);
    return -EBADF;
  }

  rc = chsize(f, size);

  hrel(h);
  unlock_buffer(params, 8);

  return rc;
}

static int sys_fstat(char *params)
{
  struct file *f;
  handle_t h;
  struct stat *buffer;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  h = *(handle_t *) params;
  buffer = *(struct stat **) (params + 4);

  f = (struct file *) hlock(h, OBJECT_FILE);
  if (!f) 
  {
    unlock_buffer(params, 8);
    return -EBADF;
  }

  if (lock_buffer(buffer, sizeof(struct stat)) < 0)
  {
    hrel(h);
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  rc = fstat(f, buffer);

  unlock_buffer(buffer, sizeof(struct stat));
  hrel(h);
  unlock_buffer(params, 8);

  return rc;
}

static int sys_stat(char *params)
{
  char *name;
  struct stat *buffer;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  name = *(char **) params;
  buffer = *(struct stat **) (params + 4);

  if (lock_string(name) < 0) 
  {
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  if (lock_buffer(buffer, sizeof(struct stat)) < 0)
  {
    unlock_string(name);
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  rc = stat(name, buffer);

  unlock_buffer(buffer, sizeof(struct stat));
  unlock_string(name);
  unlock_buffer(params, 8);

  return rc;
}

static int sys_mkdir(char *params)
{
  char *name;
  int rc;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  name = *(char **) params;

  if (lock_string(name) < 0) 
  {
    unlock_buffer(params, 4);
    return -EFAULT;
  }

  rc = mkdir(name);

  unlock_string(name);
  unlock_buffer(params, 4);

  return rc;
}

static int sys_rmdir(char *params)
{
  char *name;
  int rc;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  name = *(char **) params;

  if (lock_string(name) < 0) 
  {
    unlock_buffer(params, 4);
    return -EFAULT;
  }

  rc = rmdir(name);

  unlock_string(name);
  unlock_buffer(params, 4);

  return rc;
}

static int sys_rename(char *params)
{
  char *oldname;
  char *newname;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  oldname = *(char **) params;
  newname = *(char **) (params + 4);

  if (lock_string(oldname) < 0) 
  {
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  if (lock_string(newname) < 0) 
  {
    unlock_string(oldname);
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  rc = rename(oldname, newname);

  unlock_string(newname);
  unlock_string(oldname);
  unlock_buffer(params, 8);

  return rc;
}

static int sys_link(char *params)
{
  char *oldname;
  char *newname;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  oldname = *(char **) params;
  newname = *(char **) (params + 4);

  if (lock_string(oldname) < 0) 
  {
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  if (lock_string(newname) < 0) 
  {
    unlock_string(oldname);
    unlock_buffer(params, 8);
    return -EFAULT;
  }

  rc = link(oldname, newname);

  unlock_string(newname);
  unlock_string(oldname);
  unlock_buffer(params, 8);

  return rc;
}

static int sys_unlink(char *params)
{
  char *name;
  int rc;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  name = *(char **) params;

  if (lock_string(name) < 0) 
  {
    unlock_buffer(params, 4);
    return -EFAULT;
  }

  rc = unlink(name);

  unlock_string(name);
  unlock_buffer(params, 4);

  return rc;
}

static int sys_opendir(char *params)
{
  struct file *f;
  char *name;
  int rc;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  name = *(char **) params;

  if (lock_string(name) < 0)
  {
    unlock_buffer(params, 4);
    return -EFAULT;
  }

  rc = opendir(name, &f);
  if (rc == 0)
  {
    rc = halloc(&f->object);
    if (rc < 0) close(f);
  }

  unlock_string(name);
  unlock_buffer(params, 4);

  return rc;
}

static int sys_readdir(char *params)
{
  handle_t h;
  struct file *f;
  int rc;
  struct dirent *dirp;
  int count;

  if (lock_buffer(params, 12) < 0) return -EFAULT;

  h = *(handle_t *) params;
  dirp = *(struct dirent **) (params + 4);
  count = *(int *) (params + 8);

  f = (struct file *) hlock(h, OBJECT_FILE);
  if (f == NULL) 
  {
    unlock_buffer(params, 12);
    return -EBADF;
  }

  if (lock_buffer(dirp, count * sizeof(struct dirent)) < 0)
  {
    hrel(h);
    unlock_buffer(params, 12);
    return -EFAULT;
  }
  
  rc = readdir(f, dirp, count);
  
  hrel(h);
  unlock_buffer(params, 12);

  return rc;
}

static int sys_mmap(char *params)
{
  char *addr;
  unsigned long size;
  int type;
  int protect;
  void *retval;

  if (lock_buffer(params, 16) < 0) return -EFAULT;

  addr = *(void **) params;
  size = *(unsigned long *) (params + 4);
  type = *(int *) (params + 8);
  protect = *(int *) (params + 12);

  retval = mmap(addr, size, type, protect);

  unlock_buffer(params, 16);

  return (int) retval;
}

static int sys_munmap(char *params)
{
  char *addr;
  unsigned long size;
  int type;
  int rc;

  if (lock_buffer(params, 12) < 0) return -EFAULT;

  addr = *(void **) params;
  size = *(unsigned long *) (params + 4);
  type = *(int *) (params + 8);

  rc = munmap(addr, size, type);

  unlock_buffer(params, 12);

  return rc;
}

static int sys_mremap(char *params)
{
  char *addr;
  unsigned long oldsize;
  unsigned long newsize;
  int type;
  int protect;
  void *retval;

  if (lock_buffer(params, 20) < 0) return -EFAULT;

  addr = *(void **) params;
  oldsize = *(unsigned long *) (params + 4);
  newsize = *(unsigned long *) (params + 8);
  type = *(int *) (params + 12);
  protect = *(int *) (params + 16);

  retval = mremap(addr, oldsize, newsize, type, protect);

  unlock_buffer(params, 20);

  return (int) retval;
}

static int sys_mprotect(char *params)
{
  char *addr;
  unsigned long size;
  int protect;
  int rc;

  if (lock_buffer(params, 12) < 0) return -EFAULT;

  addr = *(void **) params;
  size = *(unsigned long *) (params + 4);
  protect = *(int *) (params + 8);

  rc = mprotect(addr, size, protect);

  unlock_buffer(params, 12);

  return rc;
}

static int sys_mlock(char *params)
{
  char *addr;
  unsigned long size;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  addr = *(void **) params;
  size = *(unsigned long *) (params + 4);

  rc = mlock(addr, size);

  unlock_buffer(params, 8);

  return rc;
}

static int sys_munlock(char *params)
{
  char *addr;
  unsigned long size;
  int rc;

  if (lock_buffer(params, 8) < 0) return -EFAULT;

  addr = *(void **) params;
  size = *(unsigned long *) (params + 4);

  rc = munlock(addr, size);

  unlock_buffer(params, 8);

  return rc;
}

static int sys_wait(char *params)
{
  return -ENOSYS;
}

static int sys_mkevent(char *params)
{
  int manual_reset;
  int initial_state;
  struct event *e;
  handle_t h;

  lock_buffer(params, 8);

  manual_reset = *(int *) params;
  initial_state = *(int *) (params + 4);

  e = (struct event *) kmalloc(sizeof(struct event));
  if (!e) 
  {
    unlock_buffer(params, 8);
    return -ENOMEM;
  }

  init_event(e, manual_reset, initial_state);

  h = halloc(&e->object);
  if (h < 0)
  {
    close_object(&e->object);
    kfree(e);
    unlock_buffer(params, 8);
    return h;
  }

  unlock_buffer(params, 8);

  return h;
}

static int sys_epulse(char *params)
{
  struct event *e;
  handle_t h;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  h = *(handle_t *) params;
  e = (struct event *) hlock(h, OBJECT_EVENT);
  if (!e) 
  {
    unlock_buffer(params, 4);
    return -EBADF;
  }

  pulse_event(e);

  hrel(h);
  unlock_buffer(params, 4);

  return 0;
}

static int sys_eset(char *params)
{
  struct event *e;
  handle_t h;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  h = *(handle_t *) params;
  e = (struct event *) hlock(h, OBJECT_EVENT);
  if (!e) 
  {
    unlock_buffer(params, 4);
    return -EBADF;
  }

  set_event(e);

  hrel(h);
  unlock_buffer(params, 4);

  return 0;
}

static int sys_ereset(char *params)
{
  struct event *e;
  handle_t h;

  if (lock_buffer(params, 4) < 0) return -EFAULT;

  h = *(handle_t *) params;
  e = (struct event *) hlock(h, OBJECT_EVENT);
  if (!e) 
  {
    unlock_buffer(params, 4);
    return -EBADF;
  }

  reset_event(e);

  hrel(h);
  unlock_buffer(params, 4);

  return 0;
}

static int sys_self(char *params)
{
  return current_thread()->self;
}

static int sys_exit(char *params)
{
  exit();
  return 0;
}

int syscall(int syscallno, char *params)
{
  //kprintf("syscall %d (%p)\n", syscallno, params);
  switch (syscallno)
  {
    case SYSCALL_NULL:     return 0;
    case SYSCALL_FORMAT:   return sys_format(params);
    case SYSCALL_MOUNT:    return sys_mount(params);
    case SYSCALL_UNMOUNT:  return sys_unmount(params);
    case SYSCALL_OPEN:     return sys_open(params);
    case SYSCALL_CLOSE:    return sys_close(params);
    case SYSCALL_FLUSH:    return sys_flush(params);
    case SYSCALL_READ:     return sys_read(params);
    case SYSCALL_WRITE:    return sys_write(params);
    case SYSCALL_TELL:     return sys_tell(params);
    case SYSCALL_LSEEK:    return sys_lseek(params);
    case SYSCALL_CHSIZE:   return sys_chsize(params);
    case SYSCALL_FSTAT:    return sys_fstat(params);
    case SYSCALL_STAT:     return sys_stat(params);
    case SYSCALL_MKDIR:    return sys_mkdir(params);
    case SYSCALL_RMDIR:    return sys_rmdir(params);
    case SYSCALL_RENAME:   return sys_rename(params);
    case SYSCALL_LINK:     return sys_link(params);
    case SYSCALL_UNLINK:   return sys_unlink(params);
    case SYSCALL_OPENDIR:  return sys_opendir(params);
    case SYSCALL_READDIR:  return sys_readdir(params);
    case SYSCALL_MMAP:     return sys_mmap(params);
    case SYSCALL_MUNMAP:   return sys_munmap(params);
    case SYSCALL_MREMAP:   return sys_mremap(params);
    case SYSCALL_MPROTECT: return sys_mprotect(params);
    case SYSCALL_MLOCK:    return sys_mlock(params);
    case SYSCALL_MUNLOCK:  return sys_munlock(params);
    case SYSCALL_WAIT:     return sys_wait(params);
    case SYSCALL_MKEVENT:  return sys_mkevent(params);
    case SYSCALL_EPULSE:   return sys_epulse(params);
    case SYSCALL_ESET:     return sys_eset(params);
    case SYSCALL_ERESET:   return sys_ereset(params);
    case SYSCALL_SELF:     return sys_self(params);
    case SYSCALL_EXIT:     return sys_exit(params);
  }

  return -ENOSYS;
}
