//
// os.c
//
// Copyright (c) 2001 Michael Ringgaard. All rights reserved.
//
// Operating system API
//

#include <os.h>
#include <string.h>

#include <os/seg.h>
#include <os/tss.h>
#include <os/syspage.h>

#include "heap.h"
#include "mod.h"

struct critsect heap_lock;
struct critsect mod_lock;

unsigned long loglevel = LOG_DEBUG | LOG_HEAP | LOG_MODULE | LOG_APITRACE | LOG_AUX;

void panic(char *msg)
{
  write(2, msg, strlen(msg));
  exit();
}

void syslog(int priority, const char *fmt,...)
{
  va_list args;
  char buffer[1024];

  if (!(priority & LOG_SUBSYS_MASK)) priority |= LOG_AUX;

  if (((unsigned) priority & LOG_LEVEL_MASK) >= (loglevel & LOG_LEVEL_MASK) && (priority & loglevel & LOG_SUBSYS_MASK) != 0)
  {
    va_start(args, fmt);
    vsprintf(buffer, fmt, args);
    va_end(args);
    write(1, buffer, strlen(buffer));
  }
}

void *malloc(size_t size)
{
  void *p;

  //syslog(LOG_MODULE | LOG_DEBUG, "malloc %d bytes\n", size);

  enter(&heap_lock);
  p = heap_alloc(size);
  leave(&heap_lock);

  //syslog(LOG_MODULE | LOG_DEBUG, "malloced %d bytes at %p\n", size, p);

  return p;
}

void *realloc(void *mem, size_t size)
{
  void *p;

  enter(&heap_lock);
  p = heap_realloc(mem, size);
  leave(&heap_lock);
  return p;
}

void *calloc(size_t num, size_t size)
{
  void *p;

  enter(&heap_lock);
  p = heap_calloc(num, size);
  leave(&heap_lock);
  return p;
}

void free(void *p)
{
  enter(&heap_lock);
  heap_free(p);
  leave(&heap_lock);
}

char *canonicalize(char *filename, char *buffer, int size)
{
  char *basename = filename;
  char *p = filename;

  while (*p)
  {
    *buffer = *p;
    if (*buffer == PS1 || *buffer == PS2) basename = buffer + 1;
    buffer++;
    p++;
  }
  *buffer = 0;
  return basename;
}

void *resolve(hmodule_t hmod, char *procname)
{
  void *addr;

  enter(&mod_lock);
  addr = get_proc_address(hmod, procname);
  leave(&mod_lock);
  return addr;
}

hmodule_t getmodule(char *name)
{
  hmodule_t hmod;

  enter(&mod_lock);
  hmod = get_module_handle(name);
  leave(&mod_lock);
  return hmod;
}

int getmodpath(hmodule_t hmod, char *buffer, int size)
{
  int rc;

  enter(&mod_lock);
  rc = get_module_filename(hmod, buffer, size);
  leave(&mod_lock);
  return rc;
}

hmodule_t load(char *name)
{
  hmodule_t hmod;

  enter(&mod_lock);
  hmod = load_module(name);
  leave(&mod_lock);
  return hmod;
}

int __stdcall start(hmodule_t hmod, void *reserved, void *reserved2)
{
  char *cmdline;
  char *pgm;
  hmodule_t hexecmod;
  int rc;

  // Set usermode segment selectors
  __asm
  {
    mov	ax, SEL_UDATA + SEL_RPL3
    mov	ds, ax
    mov	es, ax
  }

  cmdline = "sh.exe 1 2 3";

  // Initialize heap and module locks
  mkcs(&heap_lock);
  mkcs(&mod_lock);

  // Initialize module table
  init_modules(hmod, "/os");

  syscall(0, NULL);

  // Parse command line
  while (*cmdline == ' ') cmdline++;
  pgm = cmdline;
  while (*cmdline && *cmdline != ' ') cmdline++;
  if (*cmdline == ' ') 
  {
    *cmdline++ = 0;
    while (*cmdline == ' ') cmdline++;
  }

  hexecmod = load_module(pgm);
  if (hexecmod == NULL) panic("unable to load executable");
  syslog(LOG_DEBUG, "exec main(%s)\n", cmdline);

  rc = ((int (__stdcall *)(hmodule_t, char *, int)) get_entrypoint(hexecmod))(hexecmod, cmdline, 0);

  if (rc != 0) syslog(LOG_DEBUG, "Exitcode: %s\n", rc);

  exit();
  return 0;
}
